/* Copyright (c) 2004 Joshua Baker
   See the file COPYING for copying permission.
*/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "xaint.h"


static void 
xaintStreamDataRequest(XaintStream *xaintst)
{
  FILE *f;
  char buffer[10];
  size_t bytesRead;
  f = (FILE*)XaintStream_getUserData(xaintst);
  bytesRead = fread(buffer, 1, sizeof(buffer), f);
  XaintStream_feedData(xaintst, buffer, bytesRead);
}


static void 
xaintTreeDataRequest(XaintTree *xainttr)
{
  FILE *f;
  char buffer[10];
  size_t bytesRead;
  f = (FILE*)XaintTree_getUserData(xainttr);
  bytesRead = fread(buffer, 1, sizeof(buffer), f);
  XaintTree_feedData(xainttr, buffer, bytesRead);
}


static void
xaintParserDataRequest(XaintParser *xaintpr)
{
  FILE *f;
  char buffer[10];
  size_t bytesRead;
  f = (FILE*)XaintParser_getUserData(xaintpr);
  bytesRead = fread(buffer, 1, sizeof(buffer), f);
  XaintParser_feedData(xaintpr, buffer, bytesRead);
}


static void xaintParserEvent(XaintParser *xaintpr, unsigned char *data, unsigned char type)
{
  xaintpr = xaintpr;
  printf("Event [%s] type: '%c'\n", data, type);
  
  
}



static void PairEvent(XaintStream *xaintst, unsigned char *key, unsigned char *value, unsigned char* ws, unsigned char type)
{
  xaintst = xaintst;
  ws = ws;  
  printf("Pair Event [%s] \"%s\" type: '%c'\n", key, value, type);
}

static void CommentEvent(XaintStream *xaintst, unsigned char *data)
{
  xaintst;
  printf("Comment Event (%s)\n", data);
  
  
}

static void PragmaEvent(XaintStream *xaintst, unsigned char *data)
{
  xaintst;
  printf("Pragma Event (%s)\n", data);
  
  
}

static void WhitespaceEvent(XaintStream *xaintst, unsigned char *data)
{
  xaintst; data;
  //printf("Whitespace Event (%s)\n", data);
  
  
}




static void Dump(XaintElement *element, int depth)
{
  int i;
  while (element)
  {
    switch (XaintElement_getType(element))
    {
      case XAINT_TYPE_LIST:
        for (i=0;i<depth*2;i++) printf(" ");
        if (!XaintElement_isAnonymous(element))
        {
          printf("[%s]\n", XaintElement_getName(element));
        }
        for (i=0;i<depth*2;i++) printf(" ");
        printf("{\n");
        Dump(XaintElement_getChild(element), depth+1);
        for (i=0;i<depth*2;i++) printf(" ");
        printf("}\n");
        break;
      case XAINT_TYPE_STRING:
        for (i=0;i<depth*2;i++) printf(" ");
        if (!XaintElement_isAnonymous(element))
        {
          printf("[%s]%s\"%s\"\n", XaintElement_getName(element), XaintElement_getWS(element), XaintElement_getValue(element));
        }
        else
        {
          printf("\"%s\"\n", XaintElement_getValue(element));
          
        }
        break;
    }
    element = XaintElement_getNext(element);
  }
}


static XaintBool 
xaintTreeDataEvent(XaintTree *xainttr, void *data, XaintNum size, XaintBool finished)
{
  xainttr; size;
  if (!finished)
  {
    printf("%s", data);
  }
  return 1;
}


int main()
{
#if 1   //Test 1
  
  XaintTree *xainttr;
  xainttr = XaintTree_new();
  XaintTree_parseString(xainttr, (XaintChar*)"[Hello] ([Hello2] \"Um, hi...\" [Hello3] (*Comment* \"String\" *Comment 2*))");
  XaintTree_setAutoWS(xainttr, 1, (XaintChar*)"\n", '\t', 1);
  {
    XaintChar *mem;
    mem = XaintTree_saveToMemory(xainttr);
    printf("%s", mem);
    Xaint_free(mem);
  }
  XaintTree_delete(xainttr);

#elif 1   // Test 2
  
  XaintTree *xainttr;
  XaintElement *element;
  void *allocmem;

l__endless_loop:

  xainttr = XaintTree_new();
  if (!XaintTree_load(xainttr, "examples/odd.xnt"))
  {
    printf("ERROR %s!\n", XaintTree_getLastError(xainttr)==XAINT_ERR_TOK?"XAINT_ERR_TOK":
                          XaintTree_getLastError(xainttr)==XAINT_ERR_OOM?"XAINT_ERR_OOM":
                          XaintTree_getLastError(xainttr)==XAINT_ERR_FIL?"XAINT_ERR_FIL":
                          "XAINT_ERR_NIL");
    printf("Line %d, Column %d\n", XaintTree_getRowNum(xainttr), XaintTree_getColNum(xainttr));
    exit(0);
  }
  {
    XaintElement *el1, *el2, *el3;
    el1 = XaintElement_new(xainttr);
    XaintElement_setName(el1, "My New List");
    XaintElement_setType(el1, XAINT_TYPE_LIST);
    XaintElement_push  (0, el1);
    XaintElement_setProperty(el1, "%1", "Value 1");
    XaintElement_setProperty(el1, "%2", "Value 2");
    XaintElement_setProperty(el1, "%3", "Value 3");
    XaintElement_setProperty(el1, "%4", "Value 4");
    //XaintElement_delete(el1); 
  }
  XaintTree_setAutoWS(xainttr, 1, "\n", '\t', 1);
  {
    XaintChar *mem;
    mem = XaintTree_saveToMemory(xainttr);
    printf("%s", mem);
    free(mem);
  }
  XaintTree_delete(xainttr);
  //goto l__endless_loop;
  return 0;
  
#elif 1 //Test 3
  XaintStream *xaintst;
  FILE *f;
  
l__endless_loop:

  f = fopen("examples/breakfast-menu.xnt", "rb");
  
  xaintst = XaintStream_new();
  XaintStream_setUserData(xaintst, f);
  XaintStream_setDataRequest(xaintst, xaintStreamDataRequest);
  XaintStream_setPairEvent(xaintst, PairEvent);
  XaintStream_setPragmaEvent(xaintst, PragmaEvent);
  XaintStream_setWhitespaceEvent(xaintst, WhitespaceEvent);
  XaintStream_setCommentEvent(xaintst, CommentEvent);
  XaintStream_setStoreRecordNames(xaintst, 1);
  if (!XaintStream_parse(xaintst))
  {
    printf("ERROR %s!\n", 
           XaintStream_getLastError(xaintst)==XAINT_ERR_TOK?"XAINT_ERR_TOK":
           XaintStream_getLastError(xaintst)==XAINT_ERR_OOM?"XAINT_ERR_OOM":
           "XAINT_ERR_NIL");
    exit(0);
  }
  fclose(f);  
  XaintStream_delete(xaintst);
  //goto l__endless_loop;
  return 0;
#elif 1 // Test 4
  XaintParser *xaintpr;
  xaintpr = XaintParser_new();
  XaintParser_setEvent(xaintpr, xaintParserEvent);
  XaintParser_parseString(xaintpr, "[hello] (\"asdfasdfasdfasdf\")");
  XaintParser_delete(xaintpr);
  return 0;
#else  // Test 5
  XaintParser *xaintpr;
  FILE *f;
  f = fopen("examples/note.xnt", "rb");
  xaintpr = XaintParser_new();
  XaintParser_setUserData(xaintpr, f);
  XaintParser_setDataRequest(xaintpr, xaintParserDataRequest);
  XaintParser_setEvent(xaintpr, xaintParserEvent);
  if (!XaintParser_parse(xaintpr))
  {
    printf("ERROR %s!\n", 
           XaintParser_getLastError(xaintpr)==XAINT_ERR_TOK?"XAINT_ERR_TOK":
           XaintParser_getLastError(xaintpr)==XAINT_ERR_OOM?"XAINT_ERR_OOM":
           "XAINT_ERR_NIL");
    exit(0);
  }
  fclose(f);
  XaintParser_delete(xaintpr);
  return 0;
#endif
}



