/* Copyright (c) 2004 Joshua Baker
   See the file COPYING for copying permission.
*/

#ifndef __XAINT_H__
#define __XAINT_H__
#ifdef __cplusplus
extern "C" {
#endif

#ifdef XAINT_DLL
#define XAINT_API __declspec(dllexport)
#else
#define XAINT_API
#endif



/* Xaint Data Types */
typedef unsigned char XaintType;
typedef unsigned char XaintChar;
typedef unsigned long XaintBool;
typedef unsigned long XaintNum;
typedef unsigned long XaintErr;
typedef void * XaintUnknown;

/* Xaint Errors. */
enum
{
  XAINT_ERR_NIL = 0,
  XAINT_ERR_OOM = 1,
  XAINT_ERR_TOK = 2,
  XAINT_ERR_FIL = 3,
};

/* Xaint Element Types. */
enum
{
  XAINT_TYPE_LIST = '(',
  XAINT_TYPE_SLIST = '(',
  XAINT_TYPE_ELIST = ')',
  XAINT_TYPE_SNAME = '[',
  XAINT_TYPE_NAME = ']',
  XAINT_TYPE_ENAME = ']',
  XAINT_TYPE_STRING = '\"',
  XAINT_TYPE_COMMENT = '*',
  XAINT_TYPE_PRAGMA = '?',
  XAINT_TYPE_WS = ' ',
  XAINT_TYPE_TREE = 'T',  /* Used with the XaintTree Class. */
};

/* Xaint Boolean Values. */
enum
{
  XAINT_FALSE = 0,
  XAINT_TRUE = 1,
};




/*
 ***************************************************

                 XAINT_PARSER HEADER

 ***************************************************
*/






/* The XaintParser Class Type */
typedef struct XaintParser_type XaintParser;

/* Callback Events Types */
typedef void (*XaintParserEventFunc)(XaintParser *xaint, XaintChar *data, XaintType type);
typedef void (*XaintParserDataRequestFunc)(XaintParser *xaint);

/* Xaint Instance Creation/Deletion Methods */
XAINT_API XaintParser *XaintParser_new(void);
XAINT_API void XaintParser_delete(XaintParser *xaint);

/* Xaint Instance Properties */
  /* (set/get)Event*/
XAINT_API void XaintParser_setEvent(XaintParser *xaintpr, XaintParserEventFunc event);
XAINT_API XaintParserEventFunc XaintParser_getEvent(XaintParser *xaintpr);

  /* (set/get)DataRequest */
XAINT_API void XaintParser_setDataRequest(XaintParser *xaintpr, XaintParserDataRequestFunc dataRequest);
XAINT_API XaintParserDataRequestFunc XaintParser_getDataRequest(XaintParser *xaintpr);

  /* (set/get)UserData */
XAINT_API void XaintParser_setUserData(XaintParser *xaintpr, XaintUnknown userData);
XAINT_API XaintUnknown XaintParser_getUserData(XaintParser *xaintpr);

  /* (get)LastError */
XAINT_API XaintErr XaintParser_getLastError(XaintParser *xaintpr);

  /* (get)RowNum */
XAINT_API XaintNum XaintParser_getRowNum(XaintParser *xaintpr);

  /* (get)ColNum */
XAINT_API XaintNum XaintParser_getColNum(XaintParser *xaintpr);

/* Xaint Instance Methods */
XAINT_API void XaintParser_feedData(XaintParser *xaintpr, XaintUnknown data, XaintNum size);
XAINT_API XaintBool XaintParser_parse(XaintParser *xaintpr);
XAINT_API XaintBool XaintParser_parseString(XaintParser *xaintpr, XaintChar *string);


/* Xaint Class Methods */
XAINT_API void Xaint_free(void *ptr);
XAINT_API void *Xaint_malloc(XaintNum size);
XAINT_API void *Xaint_realloc(void *ptr, XaintNum size);


/* Xaint Internal Handlings. */
XAINT_API void XaintParser_INTERNAL_oom(XaintParser *xaintpr, int v);
XAINT_API void XaintParser_INTERNAL_set_lastError(XaintParser *xaintpr, XaintNum lastError);

/* Simple UTF8/UCS4 En/Decoding Utility Methods and Types */
typedef unsigned char UTF8Char;
typedef unsigned long UCS4Char;
XAINT_API UCS4Char *CreateUCS4String(UTF8Char *string);
XAINT_API UTF8Char *CreateUTF8String(UCS4Char *string);
XAINT_API void FreeUTF8String(UTF8Char *string);
XAINT_API void FreeUCS4String(UCS4Char *string);

/* Expose some functions for handling a CPP wrapper. 
   Not technically a part of the API. */
typedef void (*XaintParserOnDelete)(XaintParser *xaintpr);
XAINT_API XaintParserOnDelete XaintParser_getOnDelete(XaintParser *xaintpr);
XAINT_API void XaintParser_setOnDelete(XaintParser *xaintpr, XaintParserOnDelete OnDelete);
XAINT_API void XaintParser_setClassInstance(XaintParser *xaintpr, XaintUnknown ClassInstance);
XAINT_API XaintUnknown XaintParser_getClassInstance(XaintParser *xaintpr);



/*
 ***************************************************

                 XAINT_STREAM HEADER

 ***************************************************
*/



/* The XaintStream Class Type */
typedef struct XaintStream_type XaintStream;

/* XaintStream Callback Event Types */
typedef void (*XaintStreamPairEvent)(XaintStream *xaintst, XaintChar *key, XaintChar *value, XaintChar *ws, XaintChar type);
typedef void (*XaintStreamWhitespaceEvent)(XaintStream *xaintst, XaintChar *data);
typedef void (*XaintStreamPragmaEvent)(XaintStream *xaintst, XaintChar *data);
typedef void (*XaintStreamCommentEvent)(XaintStream *xaintst, XaintChar *data);
typedef void (*XaintStreamDataRequest)(XaintStream *xaintst);

/* XaintStream Instance Creation/Deletion Methods */
XaintStream *XaintStream_new(void);
XAINT_API void XaintStream_delete(XaintStream *xaintst);

/* XaintStream Instance Properties */
  /* (get/set)StoreRecordNames */
XAINT_API void XaintStream_setStoreRecordNames(XaintStream *xaintst, XaintBool storeRecordNames);
XAINT_API XaintBool XaintStream_getStoreRecordNames(XaintStream *xaintst);

  /* (get)RecordName */
XAINT_API XaintChar *XaintStream_getRecordName(XaintStream *xaintst);

  /* (get/set)PairEvent */
XAINT_API void XaintStream_setPairEvent(XaintStream *xaintst, XaintStreamPairEvent event);
XAINT_API XaintStreamPairEvent XaintStream_getPairEvent(XaintStream *xaintst);

  /* (get/set)WhitespaceEvent */
XAINT_API void XaintStream_setWhitespaceEvent(XaintStream *xaintst, XaintStreamWhitespaceEvent event);
XAINT_API XaintStreamWhitespaceEvent XaintStream_getWhitespaceEvent(XaintStream *xaintst);

  /* (get/set)PragmaEvent */
XAINT_API void XaintStream_setPragmaEvent(XaintStream *xaintst, XaintStreamPragmaEvent event);
XAINT_API XaintStreamPragmaEvent XaintStream_getPragmaEvent(XaintStream *xaintst);

  /* (get/set)CommentEvent */
XAINT_API void XaintStream_setCommentEvent(XaintStream *xaintst, XaintStreamCommentEvent event);
XAINT_API XaintStreamCommentEvent XaintStream_getCommentEvent(XaintStream *xaintst);

  /* (get/set)DataRequest */
XAINT_API void XaintStream_setDataRequest(XaintStream *xaintst, XaintStreamDataRequest dataRequest);
XAINT_API XaintStreamDataRequest XaintStream_getDataRequest(XaintStream *xaintst);

  /* (get/set)UserData */
XAINT_API void XaintStream_setUserData(XaintStream *xaintst, XaintUnknown userData);
XAINT_API XaintUnknown XaintStream_getUserData(XaintStream *xaintst);

  /* (get)LastError */
XAINT_API XaintErr XaintStream_getLastError(XaintStream *xaintst);

  /* (get)RowNum */
XAINT_API XaintNum XaintStream_getRowNum(XaintStream *xaintst);

  /* (get)ColNum */
XAINT_API XaintNum XaintStream_getColNum(XaintStream *xaintst);

/* XaintStream Instance Methods */
XAINT_API void XaintStream_feedData(XaintStream *xaintst, XaintUnknown data, XaintNum size);
XAINT_API XaintBool XaintStream_parse(XaintStream *xaintst);
XAINT_API XaintBool XaintStream_parseString(XaintStream *xaintst, XaintChar *string);
XAINT_API XaintParser *XaintStream_INTERNAL_get_xaintpr(XaintStream *xaintst);

/* Expose some functions for handling the CPP wrapper. 
   Not technically a part of the API. */
typedef void (*XaintStreamOnDelete)(XaintStream *xaint);
XAINT_API XaintStreamOnDelete XaintStream_getOnDelete(XaintStream *xaintst);
XAINT_API void XaintStream_setOnDelete(XaintStream *xaintst, XaintStreamOnDelete onDelete);
XAINT_API void XaintStream_setClassInstance(XaintStream *xaintst, XaintUnknown ClassInstance);
XAINT_API XaintUnknown XaintStream_getClassInstance(XaintStream *xaintst);















/*
 ***************************************************

                 XAINT_TREE HEADER

 ***************************************************
*/






/* The XaintTree Class Type */
typedef struct XaintTree_type XaintTree;

/* The XaintElement Class Type */
typedef struct XaintElement_type XaintElement;

/* XaintTree Callback Events */
typedef XaintBool (*XaintTreeDataEvent)(XaintTree *xainttr, XaintUnknown data, XaintNum size, XaintBool finished);
typedef void (*XaintTreeDataRequest)(XaintTree *xainttr);

/* XaintTree Instance Creation/Deletion */
XAINT_API XaintTree *XaintTree_new(void);
XAINT_API void XaintTree_delete(XaintTree *xainttr);

/* XaintTree Instance Properties */
  /* (get/set)UseBOM */
XAINT_API void XaintTree_setUseBOM(XaintTree *xainttr, XaintBool useBOM);
XAINT_API XaintBool XaintTree_getUseBOM(XaintTree *xainttr);

  /* (get/set)Path */
XAINT_API void XaintTree_setPath(XaintTree *xainttr, XaintChar *path);
XAINT_API XaintChar *XaintTree_getPath(XaintTree *xainttr);

  /* (get/set)UserData */
XAINT_API void XaintTree_setUserData(XaintTree *cametr, XaintUnknown userData);
XAINT_API XaintUnknown XaintTree_getUserData(XaintTree *xainttr);


  /* (get)LastError */
XAINT_API XaintErr XaintTree_getLastError(XaintTree *xainttr);

  /* (get)RowNum */
XAINT_API XaintNum XaintTree_getRowNum(XaintTree *xainttr);

  /* (get)ColNum */
XAINT_API XaintNum XaintTree_getColNum(XaintTree *xainttr);


  /* (get)Root */
XAINT_API XaintElement *XaintTree_getRoot(XaintTree *xainttr);

/* XaintTree Instance Methods */
  /* Loading/Parsing */
XAINT_API XaintBool XaintTree_load(XaintTree *xainttr, XaintChar *path);
XAINT_API XaintBool XaintTree_parse(XaintTree *xainttr);
XAINT_API XaintBool XaintTree_parseString(XaintTree *xainttr, XaintChar *string);

  /* Saving */
XAINT_API XaintBool XaintTree_save(XaintTree *xainttr);
XAINT_API XaintChar *XaintTree_saveToMemory(XaintTree *xainttr);
XAINT_API XaintBool XaintTree_saveToEvent(XaintTree *xainttr, XaintTreeDataEvent event);

  /* Misc */  
XAINT_API void XaintTree_setDataRequest(XaintTree *xainttr, XaintTreeDataRequest dataRequest);
XAINT_API void XaintTree_feedData(XaintTree *xainttr, XaintUnknown data, XaintNum size);
XAINT_API XaintBool XaintTree_setAutoWS(XaintTree *xainttr, XaintBool autoWS, XaintChar *lineBreak, XaintChar character, XaintNum count);
XAINT_API void XaintTree_deleteAllElements(XaintTree *xainttr);
XAINT_API XaintStream *XaintTree_getXaintStreamInstance(XaintTree *xainttr);


/* XaintElement Instance Creation */
XAINT_API XaintElement *XaintElement_new(XaintTree *xainttr);
XAINT_API void XaintElement_delete(XaintElement *element);

/* XaintElement Instance Properties */
  /* (get/set)Type */
XAINT_API XaintType XaintElement_getType(XaintElement *element);
XAINT_API void XaintElement_setType(XaintElement *element, XaintType type);

  /* (get/set)Value */
XAINT_API XaintChar *XaintElement_getValue(XaintElement *element);
XAINT_API void XaintElement_setValue(XaintElement *element, XaintChar *value);

  /* (get/set)Name */
XAINT_API XaintChar *XaintElement_getName(XaintElement *element);
XAINT_API void XaintElement_setName(XaintElement *element, XaintChar *name);

  /* (get/set)WS */
XAINT_API XaintChar *XaintElement_getWS(XaintElement *element);
XAINT_API void XaintElement_setWS(XaintElement *element, XaintChar *ws);


  /* (get)Tree */
XAINT_API XaintTree *XaintElement_getTree(XaintElement *element);
  /* (get)Parent */
XAINT_API XaintElement *XaintElement_getParent(XaintElement *element);
  /* (get)Next */
XAINT_API XaintElement *XaintElement_getNext(XaintElement *element);
  /* (get)Prev */
XAINT_API XaintElement *XaintElement_getPrev(XaintElement *element);
  /* (get)Child */
XAINT_API XaintElement *XaintElement_getChild(XaintElement *element);

  /* (get)NumberOfElements */
XAINT_API XaintNum XaintElement_getNumberOfElements(XaintElement *element);



  /* (get/set)Property */
XAINT_API void XaintElement_setProperty(XaintElement *element, XaintChar *name, XaintChar *value);
XAINT_API XaintChar *XaintElement_getProperty(XaintElement *element, XaintChar *name);



/* XaintElement Instance Methods */
  /* Insertion/Removal */
XAINT_API XaintBool XaintElement_push(XaintElement *parentElement, XaintElement *newElement);
XAINT_API XaintBool XaintElement_pushBefore(XaintElement *refElement, XaintElement *newElement);
XAINT_API XaintBool XaintElement_pushAfter(XaintElement *refElement, XaintElement *newElement);
XAINT_API XaintBool XaintElement_pop(XaintElement *element);

  /* Misc */
XAINT_API XaintBool XaintElement_isAnonymous(XaintElement *element);
XAINT_API XaintBool XaintElement_isEmpty(XaintElement *element);
XAINT_API XaintBool XaintElement_hasProperties(XaintElement *element);
XAINT_API XaintNum XaintElement_numberOfProperties(XaintElement *element);



/* Expose some functions for handling the CPP wrapper. 
   Not technically a part of the API. */
typedef void (*XaintTreeOnDelete)(XaintTree *xaint);
XAINT_API XaintTreeOnDelete XaintTree_getOnDelete(XaintTree *xainttr);
XAINT_API void XaintTree_setOnDelete(XaintTree *xainttr, XaintTreeOnDelete onDelete);
XAINT_API void XaintTree_setClassInstance(XaintTree *xainttr, XaintUnknown ClassInstance);
XAINT_API XaintUnknown XaintTree_getClassInstance(XaintTree *xainttr);

typedef void (*XaintElementOnDelete)(XaintElement *element);
XAINT_API XaintElementOnDelete XaintElement_getOnDelete(XaintElement *element);
XAINT_API void XaintElement_setOnDelete(XaintElement *element, XaintElementOnDelete OnDelete);
XAINT_API void XaintElement_setClassInstance(XaintElement *element, XaintUnknown ClassInstance);
XAINT_API XaintUnknown XaintElement_getClassInstance(XaintElement *element);




/* 
  GNU/Linux version convention.
  Odd number minor version for development.
  Even number minor version for stable releases.
  Micro is incremented for each release. And, set back to 0 for
  any change to the major or minor versions.
*/
#define XAINT_MAJOR_VERSION 1
#define XAINT_MINOR_VERSION 0
#define XAINT_MICRO_VERSION 1




#ifdef __cplusplus
}
#endif
#endif


