/* Copyright (c) 2004 Joshua Baker
   See the file COPYING for copying permission.
*/

#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include "xaint.h"


/* The Xaint Object Structure. */
struct XaintParser_type
{
  XaintUnknown userData;
  XaintErr lastError;
  XaintNum rowNum;
  XaintNum colNum;
  XaintUnknown privData;
  XaintParserEventFunc event;
  XaintParserDataRequestFunc dataRequest;
  XaintNum ucs;
  XaintNum lastUCS;
  XaintNum chd;
  XaintNum lookForward;
  XaintChar *buffer;
  XaintChar *bufferEnd;
  XaintChar *bufferPtr;
  jmp_buf p__eos;
  jmp_buf p__oom;
  XaintChar *memory;
  XaintNum memorySize;
  XaintNum memoryLength;
  XaintUnknown ClassInstance;
  XaintParserOnDelete OnDelete;
};


/* The number of byte that memory is alloced/realloced by. */
#define MEM_GROW  (0xFF)

/* Quick Macro's */
#define read_char() { __read_char(xaintpr); if (xaintpr->lastError) goto l__err; }
#define raw_byte (*((XaintChar*)xaintpr->bufferPtr))
#define clear_memory() (*((XaintChar*)xaintpr->memory)) = 0; xaintpr->memoryLength = 0;
#define add_memory() { __add_memory(xaintpr); if (xaintpr->lastError) goto l__err; }
#define cap_memory() ((XaintChar*)xaintpr->memory)[xaintpr->memoryLength] = 0;
#define send_event() __send_data_event(xaintpr, (XaintChar*)xaintpr->memory)
#define send_null_event() __send_data_event(xaintpr, 0)
#define eos() (longjmp(xaintpr->p__eos, 1))
#define case_space_tok case ' ':case '\n':case '\r':case '\f':case '\t':case '\v'
#define oom() ( longjmp(xaintpr->p__oom, 1) )

void Xaint_free(void *ptr) { if (ptr) free(ptr); }
void *Xaint_malloc(XaintNum size) { return malloc(size); }
void *Xaint_realloc(void *ptr, XaintNum size) { return realloc(ptr, size); }
 
static void 
__send_data_event(XaintParser *xaintpr, XaintChar *data)
{
  cap_memory();
  if (xaintpr->event)
  {
    xaintpr->event(xaintpr, data, (XaintChar)xaintpr->chd);
  }
  clear_memory();
}  
  
  
XaintParser *
XaintParser_new()
{
  XaintParser *xaintpr = (struct XaintParser_type*)Xaint_malloc(sizeof(struct XaintParser_type));
  if (!xaintpr) return 0;
  memset(xaintpr, 0, sizeof(struct XaintParser_type));
  return xaintpr;
}

void
XaintParser_delete(XaintParser *xaintpr)
{
  if (!xaintpr) return;
  if (xaintpr->OnDelete) xaintpr->OnDelete(xaintpr);
  Xaint_free(xaintpr);
}

void
XaintParser_setEvent(XaintParser *xaintpr, XaintParserEventFunc event)
{
  if (!xaintpr) return;
  xaintpr->event = event;
}

XaintParserEventFunc
XaintParser_getEvent(XaintParser *xaintpr)
{
  if (!xaintpr) return 0;
  return xaintpr->event;
}

void
XaintParser_setDataRequest(XaintParser *xaintpr, XaintParserDataRequestFunc dataRequest)
{
  if (!xaintpr) return;
  xaintpr->dataRequest = dataRequest;
}

XaintParserDataRequestFunc
XaintParser_getDataRequest(XaintParser *xaintpr)
{
  if (!xaintpr) return 0;
  return xaintpr->dataRequest;
}

void
XaintParser_setUserData(XaintParser *xaintpr, XaintUnknown userData)
{
  if (!xaintpr) return;
  xaintpr->userData = userData;
}

XaintUnknown 
XaintParser_getUserData(XaintParser *xaintpr)
{
  if (!xaintpr) return 0;
  return xaintpr->userData;
}

XaintErr 
XaintParser_getLastError(XaintParser *xaintpr)
{
  if (!xaintpr) return XAINT_ERR_NIL;
  return xaintpr->lastError;
}

XaintNum 
XaintParser_getRowNum(XaintParser *xaintpr)
{
  if (!xaintpr) return 0;
  return xaintpr->rowNum;
}

XaintNum 
XaintParser_getColNum(XaintParser *xaintpr)
{
  if (!xaintpr) return 0;
  return xaintpr->colNum;
}

void 
XaintParser_INTERNAL_oom(XaintParser *xaintpr, int v)
{
  longjmp(xaintpr->p__oom, v);
}

void 
XaintParser_INTERNAL_set_lastError(XaintParser *xaintpr, XaintNum lastError)
{
  xaintpr->lastError = lastError;
}


void
XaintParser_feedData(XaintParser *xaintpr, XaintUnknown data, XaintNum size)
{
  if (xaintpr->buffer) Xaint_free(xaintpr->buffer);
  xaintpr->buffer = (XaintChar*)Xaint_malloc(size);
  if (!xaintpr->buffer) oom();
  memcpy(xaintpr->buffer, data, size);
  xaintpr->bufferEnd = ((XaintChar*)xaintpr->buffer)+size-1;
  xaintpr->bufferPtr = xaintpr->buffer;
  if (!size) xaintpr->dataRequest = 0;
}


static void 
__add_memory(XaintParser *xaintpr)
{

  if (xaintpr->memorySize==xaintpr->memoryLength)
  {
    void *allocmem;
    xaintpr->memorySize += MEM_GROW;
    allocmem = Xaint_realloc(xaintpr->memory, xaintpr->memorySize+11);
    if (!allocmem) { xaintpr->lastError = XAINT_ERR_OOM; return; }
    xaintpr->memory = (XaintChar*)allocmem;
  }
  
  if (xaintpr->ucs>0x7F)
  {
    XaintNum i, seed;
    XaintChar data[6];
    XaintNum dataSize;
    if (xaintpr->ucs<0x80000000)
    {
      if (xaintpr->ucs>0x03FFFFFF)      dataSize = 6;
      else if (xaintpr->ucs>0x001FFFFF) dataSize = 5;
      else if (xaintpr->ucs>0x0000FFFF) dataSize = 4;
      else if (xaintpr->ucs>0x000007FF) dataSize = 3;
      else if (xaintpr->ucs>0x0000007F) dataSize = 2;
      else dataSize = 1;
      seed=(dataSize-1)*6;
      data[0] = (XaintChar)( 0xFE << (7 - dataSize) | (XaintChar)(xaintpr->ucs >> seed) );
      for (i=1;i<dataSize;i++)
      {
        seed-=6;
        data[i] = (XaintChar)( 0x80 | ((XaintChar)(xaintpr->ucs >> seed) & 0x3F) );
      }
    }
    else
    {
      //Error - invalid unicode number
      data[0] = '?';
      dataSize = 1;
    }
    for (i=0;i<dataSize;i++)
    {
      ((XaintChar*)xaintpr->memory)[xaintpr->memoryLength++] = data[i];
    }
  }
  else
  {
    ((XaintChar*)xaintpr->memory)[xaintpr->memoryLength++] = (XaintChar)xaintpr->ucs;
  }
}


static void
__read_byte(XaintParser *xaintpr)
{
  if (xaintpr->bufferPtr==xaintpr->bufferEnd)
  {
    if (!xaintpr->dataRequest) eos();
    xaintpr->dataRequest(xaintpr);
    if (!xaintpr->dataRequest) eos();
  }
  else 
  {
    xaintpr->bufferPtr++;
  }
}

static void
__read_char(XaintParser *xaintpr)
{
l__read_again:
  __read_byte(xaintpr);
  if (raw_byte>0x7F)
  {
    XaintNum ucs;
    int i, num_bytes;
    if (raw_byte>>5==0x6) num_bytes = 1;                                  
    else if (raw_byte>>4==0xE) num_bytes = 2;                             
    else if (raw_byte>>3==0x1E) num_bytes = 3;                             
    else if (raw_byte>>2==0x3E) num_bytes = 4;                             
    else if (raw_byte>>1==0x7E) num_bytes = 5;
    else num_bytes = 0;

    ucs = (raw_byte^(0xFC<<(5-num_bytes)&0xFF))<<(6*num_bytes);  
    for (i=num_bytes-1;i>=0;i--)                                              
    {                                                                         
      __read_byte(xaintpr);
      ucs|=((raw_byte^0x80)<<(6*i));
    } 
    if (ucs==0xFEFF)
    {
      // BOM Encountered (ignore)
      goto l__read_again;
    }
    xaintpr->ucs = ucs;
  }
  else
  {
    xaintpr->ucs = raw_byte;
  }
  
  
  if (xaintpr->ucs=='\n')
  {
    if (xaintpr->lastUCS!='\r')
    {
      xaintpr->rowNum++;
      xaintpr->colNum=0;
    }
  }
  else if (xaintpr->ucs=='\r')
  {
    xaintpr->rowNum++;
    xaintpr->colNum=0;
  }
  else
  {
    xaintpr->colNum++;
  }
  xaintpr->lastUCS = xaintpr->ucs;
}

XaintBool
XaintParser_parse(XaintParser *xaintpr)
{
  XaintBool retval = 1;
  
  xaintpr->rowNum = 1;
  xaintpr->colNum = 0;

  // setup the jump buffers
  if (setjmp(xaintpr->p__eos)) 
  {
    if (xaintpr->lookForward==1) 
    {
      xaintpr->lookForward = 2;
      goto l__look_forward;
    }
    else if (xaintpr->lookForward==2)
    {
      xaintpr->lookForward = 0;
      goto l__finished_parsing;
    }
    if (xaintpr->memoryLength)
    {
      cap_memory();
      send_event();
    }
    goto l__finished_parsing;
  }
  if (setjmp(xaintpr->p__oom))
  {
    xaintpr->lastError = XAINT_ERR_OOM;
    goto l__err;
  }
  
  // Setup the memory
  xaintpr->memory = (XaintChar*)Xaint_malloc(1);
  if (!xaintpr->memory) { xaintpr->lastError = XAINT_ERR_OOM; goto l__err;}
  xaintpr->memorySize = 1;
  *((XaintChar*)xaintpr->memory) = 0;
  xaintpr->memoryLength = 0;
  
l__next_item_read_char:
  xaintpr->chd = 0;
  read_char();
    
l__next_item:
  xaintpr->chd = XAINT_TYPE_WS;
  for (;;)
  {
    switch (xaintpr->ucs)
    {
    case_space_tok: break;
    default: goto l__not_space;
    }
    add_memory();
    read_char();
  }
l__not_space:
  if (xaintpr->memoryLength) send_event();
  
  /* Element */
  for (;;)
  {
    switch (xaintpr->ucs)
    {
      case XAINT_TYPE_SNAME:
        xaintpr->chd = XAINT_TYPE_NAME;
        goto l__get_odd_token_data;
      case XAINT_TYPE_COMMENT:case XAINT_TYPE_STRING:
      case XAINT_TYPE_PRAGMA:
        xaintpr->chd = xaintpr->ucs;
        l__get_odd_token_data:
        read_char();
        for (;;)
        {
          if (xaintpr->ucs==xaintpr->chd)
          {
            xaintpr->lookForward = 1;
            read_char();
            xaintpr->lookForward = 0;
            if (xaintpr->ucs!=xaintpr->chd) 
            {
              l__look_forward:
              send_event();
              if (xaintpr->lookForward==2) eos();
              xaintpr->lookForward = 0;
              goto l__next_item;
            }
          }
          add_memory();
          read_char();
        }
        case XAINT_TYPE_SLIST:
          xaintpr->chd = XAINT_TYPE_SLIST;
          send_null_event();
          goto l__next_item_read_char;
        case XAINT_TYPE_ELIST:
          xaintpr->chd = XAINT_TYPE_ELIST;
          send_null_event();
          goto l__next_item_read_char;
        default:
          xaintpr->lastError = XAINT_ERR_TOK;
          goto l__err;
    }
  }

l__err:
  retval = 0;

l__finished_parsing:
  if (xaintpr->buffer) Xaint_free(xaintpr->buffer);
  if (xaintpr->memory) Xaint_free(xaintpr->memory);
  return retval;
}


XaintParserOnDelete
XaintParser_getOnDelete(XaintParser *xaintpr)
{
  if (!xaintpr) return 0;
  return xaintpr->OnDelete;
}

void 
XaintParser_setOnDelete(XaintParser *xaintpr, XaintParserOnDelete OnDelete)
{
  if (!xaintpr) return;
  xaintpr->OnDelete = OnDelete;
}
void 
XaintParser_setClassInstance(XaintParser *xaintpr, XaintUnknown ClassInstance)
{
  if (!xaintpr) return;
  xaintpr->ClassInstance = ClassInstance;
}

XaintUnknown 
XaintParser_getClassInstance(XaintParser *xaintpr)
{
  if (!xaintpr) return 0;
  return xaintpr->ClassInstance;
}



static void
__xaintprDataRequest(XaintParser *xaintpr)
{
  if (xaintpr->privData)
  {
    XaintParser_feedData(xaintpr, xaintpr->privData, strlen((char*)(xaintpr->privData)));
    xaintpr->privData = 0;
  }
  else
  {
    XaintParser_feedData(xaintpr, 0, 0);
  }
}

XaintBool
XaintParser_parseString(XaintParser *xaintpr, XaintChar *string)
{
  xaintpr->privData = string;
  XaintParser_setDataRequest(xaintpr, __xaintprDataRequest);
  return XaintParser_parse(xaintpr);
}


// Simple UTF8/UCS4 En/Decoding Methods and types

UCS4Char *CreateUCS4String(UTF8Char *string)
{
  UCS4Char *nstring, *allocmem;
  unsigned long nstringMemSize, nstringSize;
  UTF8Char *ptr;
  UCS4Char ucs;
  int i, num_bytes;
  if (!string) return 0;

  nstring = (UCS4Char*)Xaint_malloc(sizeof(UCS4Char));
  if (!nstring) return 0;

  nstring = 0;
  nstringSize = 0;
  nstringMemSize = 0;
  
  
  ptr = string;
  while (*ptr)
  {
    
    if (*ptr>0x7F)
    {
      if (*ptr>>5==0x6) num_bytes = 1;                                  
      else if (*ptr>>4==0xE) num_bytes = 2;                             
      else if (*ptr>>3==0x1E) num_bytes = 3;                             
      else if (*ptr>>2==0x3E) num_bytes = 4;                             
      else if (*ptr>>1==0x7E) num_bytes = 5;
      else num_bytes = 0;
      
      ucs = (*ptr^(0xFC<<(5-num_bytes)&0xFF))<<(6*num_bytes);  
      for (i=num_bytes-1;i>=0;i--)                                              
      {                                                                         
        ptr++;
        if (!*ptr)
        {
          // invalid character.
          if (nstring) Xaint_free(nstring);
          return 0;
        }
        ucs|=((*ptr^0x80)<<(6*i));
      } 
    }
    else
    {
      ucs = *ptr;
    }
    if (ucs==0xFEFF)
    {
      // BOM Encountered (ignore)
    }
    else
    {
      nstringSize++;
      if (nstringSize*sizeof(UCS4Char)>=nstringMemSize)
      {
        nstringMemSize += 10*sizeof(UCS4Char); // 10 characters
        allocmem = (UCS4Char *)Xaint_realloc(nstring, nstringMemSize+sizeof(UCS4Char));
        if (!allocmem)
        {
          if (nstring) Xaint_free(nstring);
          return 0;
        }
        nstring = allocmem;
      }
      nstring[nstringSize-1] = ucs;
    }
    ptr++;
  }
  if (nstring) nstring[nstringSize] = 0;
  return nstring;
}

UTF8Char *
CreateUTF8String(UCS4Char *string)
{
  UTF8Char *nstring, *allocmem;
  unsigned long nstringMemSize, nstringSize;
  UCS4Char *ptr;
  unsigned int i, seed;
  UTF8Char data[6];
  unsigned int dataSize;
  if (!string) return 0;
  
  nstring = (UTF8Char*)Xaint_malloc(sizeof(UTF8Char));
  if (!nstring) return 0;

   
  nstring = 0;
  nstringSize = 0;
  nstringMemSize = 0;
  
  ptr = string;
  while (*ptr)
  {
   
    if (*ptr>0x7F)
    {
      if (*ptr<0x80000000)
      {
        //UNICODE
        if (*ptr>0x03FFFFFF)      dataSize = 6;
        else if (*ptr>0x001FFFFF) dataSize = 5;
        else if (*ptr>0x0000FFFF) dataSize = 4;
        else if (*ptr>0x000007FF) dataSize = 3;
        else if (*ptr>0x0000007F) dataSize = 2;
        else dataSize = 1;
        seed=(dataSize-1)*6;
        data[0] = (XaintChar) ( 0xFE << ( 7 - dataSize ) | (UTF8Char)( *ptr >> seed ) );
        for (i=1;i<dataSize;i++)
        {
          seed-=6;
          data[i] = (XaintChar) ( 0x80|((UTF8Char)(*ptr >> seed)&0x3F) );
        }
      }
      else
      {
        // Invalid Unicode Character
        if (nstring) Xaint_free(nstring);
        return 0;
      }
    }
    else
    {
      //US-ASCII
      dataSize = 1;
      data[0] = (UTF8Char)*ptr;
    }
   
    nstringSize += dataSize;
    if (nstringSize>=nstringMemSize)
    {
      nstringMemSize += 60; // 10-60 characters
      allocmem = (UTF8Char *)Xaint_realloc(nstring, nstringMemSize+1);
      if (!allocmem)
      {
        if (nstring) Xaint_free(nstring);
        return 0;
      }
      nstring = allocmem;
    }
    memcpy(nstring+nstringSize-dataSize, data, dataSize);
    ptr++;
  }
  if (nstring) nstring[nstringSize] = 0;
  return nstring;
}



void 
FreeUTF8String(UTF8Char *string)
{
  if (string) Xaint_free(string);
}
void 
FreeUCS4String(UCS4Char *string)
{
  if (string) Xaint_free(string);
}


