/* Copyright (c) 2004 Joshua Baker
   See the file COPYING for copying permission.
*/

#include <stdio.h>
#include <setjmp.h>
#include <string.h>
#include "xaint.h"

#define case_space_tok case ' ':case '\n':case '\r':case '\f':case '\t':case '\v'
  
struct
XaintElement_type
{
  XaintType type;
  XaintTree *tree;
  XaintChar *name;  
  XaintChar *value;
  XaintChar *ws;
  XaintElement *parent;
  XaintElement *next;
  XaintElement *prev;
  XaintElement *child;
  XaintUnknown ClassInstance;
  XaintElementOnDelete OnDelete;
};

struct 
XaintTree_type
{
  XaintElement tree;
  XaintUnknown userData;
  XaintStream *xaintst;
  XaintChar *path;
  XaintChar *memory;
  XaintNum memorySize;
  XaintTreeDataRequest dataRequest;
  XaintTreeDataEvent dataEvent;
  XaintElement *currentList;
  XaintBool autoWS;
  XaintChar *autoWSMemory;
  XaintChar autoWSChar;
  XaintNum autoWSCount;
  XaintChar *autoWSLineBreak;
  XaintBool useBOM;
  XaintUnknown privData;
  XaintUnknown ClassInstance;
  XaintTreeOnDelete OnDelete;
};


XaintBool 
XaintTree_setAutoWS(XaintTree *xainttr, XaintBool autoWS, XaintChar *lineBreak,
                    XaintChar character, XaintNum count)
{
  XaintChar *ptr;
  if (!xainttr) return 0;

  if (!autoWS)
  {
    xainttr->autoWS = 0;
    return 1;
  }

  if (!lineBreak) lineBreak = (XaintChar*)"\r\n";
  ptr = lineBreak;
  while (*ptr)
  {
    switch (*ptr) { case_space_tok: break; default: return 0; }
    ptr++;
  }
        
  switch (character) { case_space_tok: break; default: return 0; }
  
  if (xainttr->autoWSLineBreak) Xaint_free(xainttr->autoWSLineBreak);
  xainttr->autoWSLineBreak = (XaintChar*)Xaint_malloc(strlen((char*)lineBreak)+1);
  if (!xainttr->autoWSLineBreak) return 0;
  strcpy((char*)xainttr->autoWSLineBreak, (char*)lineBreak);
  xainttr->autoWS = autoWS;
  xainttr->autoWSChar = character;
  xainttr->autoWSCount = count;
  return 1;
}


XaintTree 
*XaintTree_new()
{
  XaintTree *xainttr;
  xainttr = (struct XaintTree_type*)Xaint_malloc(sizeof(struct XaintTree_type));
  if (!xainttr) 
  {
    return 0;
  }
  memset(xainttr, 0, sizeof(struct XaintTree_type));
  xainttr->xaintst = XaintStream_new();
  if (!xainttr->xaintst) 
  {
    XaintTree_delete(xainttr);
    return 0;
  }
  XaintStream_setUserData(xainttr->xaintst, xainttr);
  XaintTree_setUseBOM(xainttr, XAINT_TRUE);
  XaintElement_setType(&(xainttr->tree), XAINT_TYPE_TREE);
  
  return xainttr;
}


void 
XaintTree_setUseBOM(XaintTree *xainttr, XaintBool useBOM)
{
  if (!xainttr) return;
  xainttr->useBOM = useBOM;
}

XaintBool 
XaintTree_getUseBOM(XaintTree *xainttr)
{
  if (!xainttr) return 0;
  return xainttr->useBOM; 
}

void 
XaintTree_setDataRequest(XaintTree *xainttr, 
                           XaintTreeDataRequest dataRequest)
{
  if (!xainttr) return;
  xainttr->dataRequest = dataRequest;
}



static void
__DataRequest(XaintStream *xaintst)
{
  XaintTree *xainttr;
  xainttr = (XaintTree*)XaintStream_getUserData(xaintst);
  xainttr->dataRequest(xainttr);
}

void  
XaintTree_feedData(XaintTree *xainttr, 
                     void *data, 
                     XaintNum size)
{
  XaintStream_feedData(xainttr->xaintst, data, size);
}


static XaintBool
__convertTypeStringToMemory(XaintTree *xainttr, XaintChar *str, XaintType type)
{
  XaintChar *ptr;
  XaintNum i;
  void *allocmem;
  if (!xainttr||!str) return 0;
  
  if (xainttr->memory) Xaint_free(xainttr->memory);
  xainttr->memorySize = strlen((char*)str)+10;
  xainttr->memory = (XaintChar*)Xaint_malloc(xainttr->memorySize);
  
  if (!xainttr->memory) return 0;
  
  ptr = str;
  i = 0;
  while (*ptr)
  {
    if (*ptr==type)
    {
      xainttr->memorySize++;
      allocmem = (XaintChar*)Xaint_realloc(xainttr->memory, xainttr->memorySize);
      if (!allocmem)
      {
        Xaint_free(xainttr->memory);
        return 0;
      }
      xainttr->memory = (XaintChar*)allocmem;
      xainttr->memory[i++] = *ptr;
      ;
    }
    xainttr->memory[i++] = *ptr;
    ptr++;
  }
  xainttr->memory[i++] = 0;
  return 1; 
}

static XaintBool
__evDump(XaintTree *xainttr, XaintElement *element, int depth)
{
  XaintNum autoWSSize;
  if (xainttr->autoWS)
  {
    autoWSSize = depth*xainttr->autoWSCount;
    if (xainttr->autoWSMemory) Xaint_free(xainttr->autoWSMemory);
    xainttr->autoWSMemory = (XaintChar*)Xaint_malloc(autoWSSize+1);
    if (!xainttr->autoWSMemory) return 0;
    memset(xainttr->autoWSMemory, xainttr->autoWSChar, autoWSSize);
    xainttr->autoWSMemory[autoWSSize] = 0;
  }
  else
  {
    autoWSSize = 0;  
  }
  
  while (element)
  {
    switch (element->type)
    {
      case XAINT_TYPE_WS:
        if (!xainttr->autoWS)
        {
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)element->value, strlen((char*)element->value), 0)) return 0;
        }
        break;
      case XAINT_TYPE_LIST:
      {
        if (xainttr->autoWS)
        {
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->autoWSMemory, autoWSSize, 0)) return 0;  
        }
        if (element->name)
        {
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)"[", 1, 0)) return 0;
          if (!__convertTypeStringToMemory(xainttr, element->name, XAINT_TYPE_NAME))
          {
            return 0;
          }
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->memory, strlen((char*)xainttr->memory), 0)) return 0;
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)"]", 1, 0)) return 0;
        }
        if (xainttr->autoWS)
        {
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->autoWSLineBreak, strlen((char*)xainttr->autoWSLineBreak), 0)) return 0;
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->autoWSMemory, autoWSSize, 0)) return 0;
        }
        else
        {
          if (element->ws)
          {
            if (!xainttr->dataEvent(xainttr, (XaintUnknown)element->ws, strlen((char*)element->ws), 0)) return 0;
          }
        }
        if (!xainttr->dataEvent(xainttr, (XaintUnknown)"(", 1, 0)) return 0;
        if (xainttr->autoWS)
        {
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->autoWSLineBreak, strlen((char*)xainttr->autoWSLineBreak), 0)) return 0;
        }
        if (!__evDump(xainttr, element->child, depth+1))
        {
          return 0;
        }
        if (xainttr->autoWS)
        {
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->autoWSMemory, autoWSSize, 0)) return 0;  
        }
        if (!xainttr->dataEvent(xainttr, (XaintUnknown)")", 1, 0)) return 0;
        if (xainttr->autoWS)
        {
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->autoWSLineBreak, strlen((char*)xainttr->autoWSLineBreak), 0)) return 0;
        }
        break;
      }
      case XAINT_TYPE_STRING:
      {
        if (xainttr->autoWS)
        {
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->autoWSMemory, autoWSSize, 0)) return 0;  
        }
        if (XaintElement_isEmpty(element))
        {
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)"[", 1, 0)) return 0;
          if (!__convertTypeStringToMemory(xainttr, element->name, XAINT_TYPE_NAME))
          {
            return 0;
          }
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->memory, strlen((char*)xainttr->memory), 0)) return 0;
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)"]", 1, 0)) return 0;
          if (!xainttr->autoWS&&element->ws)
          {
            if (!xainttr->dataEvent(xainttr, (XaintUnknown)element->ws, strlen((char*)element->ws), 0)) return 0;
          }
        }
        else if (XaintElement_isAnonymous(element))
        {
          if (xainttr->autoWS&&strlen((char*)xainttr->autoWSLineBreak)==0&&element!=element->tree->tree.child)
          //if (xainttr->autoWS)
          {
            if (!xainttr->dataEvent(xainttr, (XaintUnknown)" ", 1, 0)) return 0;
          }
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)"\"", 1, 0)) return 0;
          if (!__convertTypeStringToMemory(xainttr, element->value, XAINT_TYPE_STRING))
          {
            return 0;
          }
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->memory, strlen((char*)xainttr->memory), 0)) return 0;
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)"\"", 1, 0)) return 0;
        }
        else
        {
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)"[", 1, 0)) return 0;
          if (!__convertTypeStringToMemory(xainttr, element->name, XAINT_TYPE_NAME))
          {
            return 0;
          }
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->memory, strlen((char*)xainttr->memory), 0)) return 0;
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)"]", 1, 0)) return 0;
          if (xainttr->autoWS)
          {
            if (!xainttr->dataEvent(xainttr, (XaintUnknown)" ", 1, 0)) return 0;
          }
          else
          {
            if (element->ws)
            {
              if (!xainttr->dataEvent(xainttr, (XaintUnknown)element->ws, strlen((char*)element->ws), 0)) return 0;
            }
          }
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)"\"", 1, 0)) return 0;
          if (!__convertTypeStringToMemory(xainttr, element->value, XAINT_TYPE_STRING))
          {
            return 0;
          }
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->memory, strlen((char*)xainttr->memory), 0)) return 0;
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)"\"", 1, 0)) return 0;
        }
        if (xainttr->autoWS)
        {
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->autoWSLineBreak, strlen((char*)xainttr->autoWSLineBreak), 0)) return 0;
        }
        break;
      }
      case XAINT_TYPE_COMMENT:
      {
        if (xainttr->autoWS)
        {
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->autoWSMemory, autoWSSize, 0)) return 0;  
        }
        if (!xainttr->dataEvent(xainttr, (XaintUnknown)"*", 1, 0)) return 0;
        if (!__convertTypeStringToMemory(xainttr, element->value, XAINT_TYPE_COMMENT))
        {
          return 0;
        }
        if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->memory, strlen((char*)xainttr->memory), 0)) return 0;
        if (!xainttr->dataEvent(xainttr, (XaintUnknown)"*", 1, 0)) return 0;
        if (xainttr->autoWS)
        {
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->autoWSLineBreak, strlen((char*)xainttr->autoWSLineBreak), 0)) return 0;
        }
        break;
      }
      case XAINT_TYPE_PRAGMA:
      {
        if (xainttr->autoWS)
        {
          if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->autoWSMemory, autoWSSize, 0)) return 0;  
        }
        if (!xainttr->dataEvent(xainttr, (XaintUnknown)"?", 1, 0)) return 0;
        if (!__convertTypeStringToMemory(xainttr, element->value, XAINT_TYPE_PRAGMA))
        {
          return 0;
        }
        if (!xainttr->dataEvent(xainttr, (XaintUnknown)xainttr->memory, strlen((char*)xainttr->memory), 0)) return 0;
        if (!xainttr->dataEvent(xainttr, (XaintUnknown)"?", 1, 0)) return 0;
        if (xainttr->autoWS)
        {
          if (!xainttr->dataEvent(xainttr, xainttr->autoWSLineBreak, strlen((char*)xainttr->autoWSLineBreak), 0)) return 0;
        }
        break;
      }
    }
    element = element->next;
  }
  return 1;
}

XaintBool 
XaintTree_saveToEvent(XaintTree *xainttr, XaintTreeDataEvent event)
{
  if (!xainttr||!event||!xainttr->tree.child) return 0;
  xainttr->dataEvent = event;
  __evDump(xainttr, xainttr->tree.child, 0);
  xainttr->dataEvent(xainttr, 0, 0, 1);
  return 1;
}


static XaintBool
__saveFileEvent(XaintTree *xainttr, void *data, XaintNum size, XaintBool finished)
{
  if (finished) return 1;
  if (fwrite(data, 1, size, (FILE*)xainttr->privData)!=size) return 0;
  return 1;
}

struct SaveMem_type
{
  XaintChar *memory;
  XaintNum size;
};

static XaintBool
__saveMemEvent(XaintTree *xainttr, void *data, XaintNum size, XaintBool finished)
{
  void *allocmem;
  struct SaveMem_type *mem;
  mem = (struct SaveMem_type*)(xainttr->privData);
  if (finished) return 1;
  mem->size+=size;
  allocmem = Xaint_realloc(mem->memory, mem->size+5);
  if (!allocmem) return 0;
  mem->memory = (XaintChar*)allocmem;
  memcpy(mem->memory+mem->size-size, data, size);
  return 1;
}


XaintChar *
XaintTree_saveToMemory(XaintTree *xainttr)
{
  struct SaveMem_type mem = {0,0};
  if (!xainttr) return 0;
  xainttr->privData = &mem;
  if (!XaintTree_saveToEvent(xainttr, __saveMemEvent))
  {
    if (mem.memory) Xaint_free(mem.memory);
    return 0;
  }
  if (mem.memory) *((XaintChar*)(mem.memory+mem.size)) = 0; 
  return (XaintChar*)mem.memory;
}

XaintBool
XaintTree_save(XaintTree *xainttr)
{
  FILE *f;
  if (!xainttr||!xainttr->path) return 0;
  f = fopen((char*)xainttr->path, "wb+");
  if (!f) return 0;
  xainttr->privData = f;
  if (!XaintTree_saveToEvent(xainttr, __saveFileEvent))
  {
    fclose(f);
    return 0;
  }
  fclose(f);
  return 1;
}

void
XaintTree_setPath(XaintTree *xainttr, XaintChar *path)
{
  if (!xainttr) return;
  if (xainttr->path) Xaint_free(xainttr->path);
  if (path)
  {
    XaintNum plen = strlen((char*)path)+1;
    xainttr->path = (XaintChar*)Xaint_malloc(plen);
    if (!xainttr->path) return;
    memcpy(xainttr->path, path, plen);
  }
  else xainttr->path = 0;
}

XaintChar *
XaintTree_getPath(XaintTree *xainttr)
{
  if (!xainttr) return 0;
  return xainttr->path;
}

void 
XaintTree_delete(XaintTree *xainttr)
{
  if (!xainttr) return;
  if (xainttr->OnDelete) xainttr->OnDelete(xainttr);
  if (xainttr->xaintst) XaintStream_delete(xainttr->xaintst);
  XaintTree_deleteAllElements(xainttr);
  if (xainttr->memory) Xaint_free(xainttr->memory);
  if (xainttr->path) Xaint_free(xainttr->path);
  if (xainttr->autoWSMemory) Xaint_free(xainttr->autoWSMemory);
  if (xainttr->autoWSLineBreak) Xaint_free(xainttr->autoWSLineBreak);
  Xaint_free(xainttr);
}

void 
XaintTree_setUserData(XaintTree *xainttr, void *userData)
{
  if (!xainttr) return;
  xainttr->userData = userData;
}

void * 
XaintTree_getUserData(XaintTree *xainttr)
{
  if (!xainttr) return 0;
  return xainttr->userData;
}


XaintElement *
XaintElement_new(XaintTree *xainttr)
{
  XaintElement *element;
  element = (struct XaintElement_type*)Xaint_malloc(sizeof(struct XaintElement_type));
  if (!element) return 0;
  memset(element, 0, sizeof(struct XaintElement_type));
  element->tree = xainttr;
  return element;
}



XaintElement *
XaintElement_getParent(XaintElement *element)
{
  if (!element) return 0;
  return element->parent;
}
XaintElement *
XaintElement_getNext(XaintElement *element)
{
  if (!element) return 0;
  return element->next;

}

XaintElement *
XaintElement_getPrev(XaintElement *element)
{
  if (!element) return 0;
  return element->prev;
}
XaintElement *
XaintElement_getChild(XaintElement *element)
{
  if (!element) return 0;
  return element->child;

}




XaintBool
XaintElement_isEmpty(XaintElement *element)
{
  if (!element) return 0;
  return (element->value==0);
}

XaintBool 
XaintElement_isAnonymous(XaintElement *element)
{
  if (!element) return 0;
  return (element->name==0);
}

void
XaintElement_setType(XaintElement *element, XaintType type)
{
  if (!element) return;
  element->type = type;
}

XaintType
XaintElement_getType(XaintElement *element)
{
  if (!element) return 0;
  return element->type;
}


void 
XaintElement_setName(XaintElement *element, XaintChar *name)
{
  if (!element) return;
  if (element->name) Xaint_free(element->name);
  
  if (name)
  {
    XaintNum nlen = strlen((char*)name)+1;
    element->name = (XaintChar*)Xaint_malloc(nlen);
    if (!element->name) return;
    memcpy(element->name, name, nlen);
  }
  else element->name = 0;
}

XaintChar *
XaintElement_getName(XaintElement *element)
{
  if (!element) return 0;
  return element->name;
}


void 
XaintTree_deleteAllElements(XaintTree *xainttr)
{
  if (!xainttr) return;
  while (xainttr->tree.child) 
  {
    XaintElement_delete(xainttr->tree.child);
  }
  xainttr->tree.child = 0;
}

void 
XaintElement_setValue(XaintElement *element, XaintChar *value)
{
  
  if (!element) return;
  if (element->value) Xaint_free(element->value);
  if (value)
  {
    XaintNum vlen = strlen((char*)value)+1;
    element->value = (XaintChar*)Xaint_malloc(vlen);
    memcpy(element->value, value, vlen);
  }
  else
  {
    element->value = 0;
  }
}

XaintChar *
XaintElement_getValue(XaintElement *element)
{
  if (!element) return 0;
  return element->value;
}

void 
XaintElement_setWS(XaintElement *element, XaintChar *ws)
{
  if (!element) return;
  if (element->ws) Xaint_free(element->ws);
  if (ws)
  {
    XaintNum wlen = strlen((char*)ws)+1;
    element->ws = (XaintChar*)Xaint_malloc(wlen);
    memcpy(element->ws, ws, wlen);
  }
  else
  {
    element->ws = 0;
  }
}

XaintChar *
XaintElement_getWS(XaintElement *element)
{
  if (!element) return 0;
  return element->ws;
}

void
XaintElement_delete(XaintElement *element)
{
  XaintElement *child, *child_next;
  XaintTree *xainttr;
  
  if (!element) return;
  xainttr = element->tree;
  
  XaintElement_pop(element);

  child = element->child;
  while (child)
  {
    child_next = child->next;
    XaintElement_delete(child);
    child = child_next;
  }
  
  if (element->name)  Xaint_free(element->name);
  if (element->value) Xaint_free(element->value);
  if (element->ws)    Xaint_free(element->ws);
  Xaint_free(element);

  return;
}

XaintBool
XaintElement_pop(XaintElement *element)
{
  if (!element) return 0;
  
  if (element->parent==0)
  {
    if (element->tree->tree.child==element)
    {
      /* root element */
      if (element->next)
      {
        element->next->prev = element->prev;
        element->next->parent = element->parent;
        element->tree->tree.child = element->next;
      }
      else
      {
        element->tree->tree.child = 0;
      }
    }
    else
    {
      if (element->next) element->next->prev = element->prev;
      if (element->prev) element->prev->next = element->next;
    }
  }
  else if (element->parent!=0)
  {
    if (element->parent==element->parent->child)
    {
      /* first child element */
      if (element->next)
      {
        element->next->prev = 0;
        element->parent->child = element->next;
      }
      else
      {
        element->parent->child = 0;
      }
    }
    else
    {
      if (element->next) element->next->prev = element->prev;
      if (element->prev) element->prev->next = element->next;
    }
  }
  element->prev = element->next = element->parent = 0;
  return 1;
}

XaintBool 
XaintElement_pushBefore(XaintElement *refElement, XaintElement *newElement)
{
  if (!newElement) return 0;
  if (!refElement) return XaintElement_push(0, newElement);
  if (refElement->tree!=newElement->tree) return 0;

  XaintElement_pop(newElement);
  newElement->parent = refElement->parent;
  if (refElement->prev)
  {
    refElement->prev->next = newElement;
  }
  newElement->prev = refElement->prev;
  refElement->prev = newElement;
  newElement->next = refElement;
  if (refElement==refElement->tree->tree.child)
  {
    refElement->tree->tree.child = newElement;
  }
  return 1;
}

XaintBool 
XaintElement_pushAfter(XaintElement *refElement, XaintElement *newElement)
{
  if (!newElement) return 0;
  if (!refElement) return XaintElement_push(0, newElement);
  if (refElement->tree!=newElement->tree) return 0;
  
  XaintElement_pop(newElement);
  newElement->parent = refElement->parent;
  newElement->prev = refElement->prev;
  newElement->next = refElement->next;
  if (refElement->next)
  {
    refElement->next->prev = newElement;
  }
  refElement->next = newElement;
  return 1;
}


XaintTree *
XaintElement_getTree(XaintElement *element)
{
  if (!element) return 0;
  return element->tree;
}



/*
XaintBool 
XaintElement_hasElements(XaintElement *element)
{
  if (!element) return 0;
  return element->child!=0;
}
*/

XaintNum 
XaintElement_getNumberOfElements(XaintElement *element)
{
  XaintElement *child;
  XaintNum num = 0;
  if (!element) return 0;
  child = element->child;
  while (child)
  {
    num++;
    child = child->next;
  }
  return num;
}

XaintElement *
XaintTree_getRoot(XaintTree *xainttr)
{
  if (!xainttr) return 0;
  return &(xainttr->tree);
}




XaintBool 
XaintElement_hasProperties(XaintElement *element)
{
  XaintElement *child;
  child = element->child;
  while (child)
  {
    if (child->type==XAINT_TYPE_STRING) return 1;
    child = child->next;
  }
  return 0;
}

XaintNum 
XaintElement_numberOfProperties(XaintElement *element)
{
  XaintElement *child;
  XaintNum num = 0;
  child = element->child;
  while (child)
  {
    if (child->type==XAINT_TYPE_STRING) num++;
    child = child->next;
  }
  return num;
}

void 
XaintElement_setProperty(XaintElement *element, XaintChar *name, XaintChar *value)
{
  XaintElement *child;
  if (!element||!name) return;
  child = element->child;
  while (child)
  {
    if (child->type==XAINT_TYPE_STRING)
    {
      if (!strcmp((char*)name, (char*)child->name))
      {
        XaintElement_setValue(child, value);
        return;
      }
    }
    child = child->next;
  }
  child = XaintElement_new(element->tree);
  if (!child) return;
  XaintElement_setType(child, XAINT_TYPE_STRING);
  if (name)
  {
    XaintElement_setName(child, name);
    if (!child->name) goto l__oom;
  }
  if (value)
  {
    XaintElement_setValue(child, value);
    if (!child->value) goto l__oom;
  }

  XaintElement_setWS(child, (XaintChar*)" ");
  if (!child->ws) goto l__oom;

  XaintElement_push(element, child);
  return;
  
l__oom:
  XaintElement_delete(child);
  return;
}
XaintChar *
XaintElement_getProperty(XaintElement *element, XaintChar *name)
{
  XaintElement *child;
  if (!element||!name) return 0;
  child = element->child;
  while (child)
  {
    if (child->type==XAINT_TYPE_STRING)
    {
      if (!strcmp((char*)name, (char*)child->name))
      {
        return child->value;
      }
    }
    child = child->next;
  }
  return 0; 
}



XaintBool
XaintElement_push(XaintElement *parentElement, XaintElement *element)
{
  XaintElement *prev;
  if (!element) return 0;
  
  XaintElement_pop(element);
  
  element->parent = parentElement;
  if (parentElement==0)
  {
    if (element->tree->tree.child==0)
    {
      element->tree->tree.child = element;
      return 1;
    }
    prev = element->tree->tree.child;
  }
  else if (!parentElement->child)
  {
    parentElement->child = element;
    return 1;
  }
  else
  {
    prev = parentElement->child;
  }
  while (prev->next)
  {
    prev = prev->next;
  }
  element->next = prev->next;
  prev->next = element;
  element->prev = prev;
  if (element->next) element->next->prev = element;  
  return 1;
}

static void 
__PairEvent(XaintStream *xaintst, 
          XaintChar *name, XaintChar *value, 
          XaintChar* ws, XaintType type)
{
  XaintTree *xainttr;
  XaintElement *element;
 
  xainttr = (XaintTree*)XaintStream_getUserData(xaintst);
  
  switch (type)
  {
    case XAINT_TYPE_ELIST:
      if (xainttr->currentList) xainttr->currentList = xainttr->currentList->parent;
      return;
  }
  if (type == XAINT_TYPE_NAME) type = XAINT_TYPE_STRING;
  element = XaintElement_new(xainttr);
  if (!element) goto l__oom;
  
  XaintElement_setType(element, type);
  if (name)
  {
    XaintElement_setName(element, name);
    if (!name) goto l__oom;
  }
  if (value)
  {
    XaintElement_setValue(element, value);
    if (!element->value) goto l__oom;
  }
  if (ws)
  {
    XaintElement_setWS(element, ws);
    if (!ws) goto l__oom;
  }

  XaintElement_push(xainttr->currentList, element);
  switch (type)
  {
    case XAINT_TYPE_SLIST:
      xainttr->currentList = element;
      break;
    default:
      break;
  }
  return;
l__oom:
  if (element) XaintElement_delete(element);
  XaintParser_INTERNAL_oom(XaintStream_INTERNAL_get_xaintpr(xainttr->xaintst), 1);
}

static void 
__CommentEvent(XaintStream *xaintst, unsigned char *data)
{
  __PairEvent(xaintst, 0, data, 0, XAINT_TYPE_COMMENT);
}

static void 
__PragmaEvent(XaintStream *xaintst, unsigned char *data)
{
  __PairEvent(xaintst, 0, data, 0, XAINT_TYPE_PRAGMA);
}

static void 
__WhitespaceEvent(XaintStream *xaintst, unsigned char *data)
{
  __PairEvent(xaintst, 0, data, 0, XAINT_TYPE_WS);
}


XaintErr 
XaintTree_getLastError(XaintTree *xainttr)
{
  if (!xainttr) return XAINT_ERR_NIL;
  return XaintStream_getLastError(xainttr->xaintst);
}


XaintNum
XaintTree_getRowNum(XaintTree *xainttr)
{
  if (!xainttr) return 0;
  return XaintStream_getRowNum(xainttr->xaintst);
}

XaintNum
XaintTree_getColNum(XaintTree *xainttr)
{
  if (!xainttr) return 0;
  return XaintStream_getColNum(xainttr->xaintst);
}

XaintBool
XaintTree_parse(XaintTree *xainttr)
{
  XaintBool retval;
  if (!xainttr) return 0;
  XaintStream_setDataRequest(xainttr->xaintst, __DataRequest);
  XaintStream_setPairEvent(xainttr->xaintst, __PairEvent);
  XaintStream_setPragmaEvent(xainttr->xaintst, __PragmaEvent);
  XaintStream_setWhitespaceEvent(xainttr->xaintst, __WhitespaceEvent);
  XaintStream_setCommentEvent(xainttr->xaintst, __CommentEvent);
  retval = XaintStream_parse(xainttr->xaintst);
  if (xainttr->memory) Xaint_free(xainttr->memory);
  xainttr->memory = 0;
  return retval;
}


static void 
__xaintTreeDataRequest_String(XaintTree *xainttr)
{
  if (xainttr->privData)
  {
    XaintTree_feedData(xainttr, xainttr->privData, strlen((char*)(xainttr->privData)));
    xainttr->privData = 0;
  }
  else
  {
    XaintTree_feedData(xainttr, 0, 0);
  }
}


XaintBool
XaintTree_parseString(XaintTree *xainttr, XaintChar *string)
{
  if (!xainttr) return 0;
  XaintTree_deleteAllElements(xainttr);
  xainttr->privData = string;
  XaintTree_setDataRequest(xainttr, __xaintTreeDataRequest_String);
  if (XaintTree_parse(xainttr))
  {
    return 1;
  }
  else
  {
    XaintTree_deleteAllElements(xainttr);
    return 0;
  }
}



static void 
__xaintTreeDataRequest(XaintTree *xainttr)
{
  FILE *f;
  char buffer[BUFSIZ];
  size_t bytesRead;
  f = (FILE*)xainttr->privData;
  bytesRead = fread(buffer, 1, sizeof(buffer), f);
  XaintTree_feedData(xainttr, buffer, bytesRead);
}


XaintStream *
XaintTree_getXaintStreamInstance(XaintTree *xainttr)
{
  if (!xainttr) return 0;
  return xainttr->xaintst;
}

XaintTreeOnDelete
XaintTree_getOnDelete(XaintTree *xainttr)
{
  if (!xainttr) return 0;
  return xainttr->OnDelete;
}

void 
XaintTree_setOnDelete(XaintTree *xainttr, XaintTreeOnDelete OnDelete)
{
  if (!xainttr) return;
  xainttr->OnDelete = OnDelete;
}
void 
XaintTree_setClassInstance(XaintTree *xainttr, void *ClassInstance)
{
  if (!xainttr) return;
  xainttr->ClassInstance = ClassInstance;
}
void *
XaintTree_getClassInstance(XaintTree *xainttr)
{
  if (!xainttr) return 0;
  return xainttr->ClassInstance;
}

XaintElementOnDelete
XaintElement_getOnDelete(XaintElement *element)
{
  if (!element) return 0;
  return element->OnDelete;
}

void 
XaintElement_setOnDelete(XaintElement *element, XaintElementOnDelete OnDelete)
{
  if (!element) return;
  element->OnDelete = OnDelete;
}
void 
XaintElement_setClassInstance(XaintElement *element, void *ClassInstance)
{
  if (!element) return;
  element->ClassInstance = ClassInstance;
}
void *
XaintElement_getClassInstance(XaintElement *element)
{
  if (!element) return 0;
  return element->ClassInstance;
}




XaintBool
XaintTree_load(XaintTree *xainttr, XaintChar *path)
{
  FILE *f;
  if (!xainttr) return 0;
  XaintTree_deleteAllElements(xainttr);
  f = fopen((char*)path, "rb");
  if (!f)
  {
    XaintParser_INTERNAL_set_lastError(XaintStream_INTERNAL_get_xaintpr(xainttr->xaintst), XAINT_ERR_FIL);
    return 0;
  }
  xainttr->privData = f;
  XaintTree_setDataRequest(xainttr, __xaintTreeDataRequest);
  if (XaintTree_parse(xainttr))
  {
    if (f) fclose(f);
    return 1;
  }
  else
  {
    XaintTree_deleteAllElements(xainttr);
    if (f) fclose(f);
    return 0;
  }
}
